/*
libhideprinters.c: 	Hides printers from applications
Copyright (C) 2005 Revolution Linux inc
Jean-Michel Dault	jmdault@revolutionlinux.com
Julien Desfossez	julien.desfossez@revolutionlinux.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <link.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <cups/http.h>
#include <cups/ipp.h>
#include <cups/cups.h>
#include <glib.h>
#include <glib/glist.h>

#define PRT_LEN 128	/* from printers.conf */
#define BUF_LEN 1024

/* From ipp_constants.h */
#define CUPS_GET_PRINTERS 0x4002

void logit(char *msg, ...)
{
	va_list fmtargs;
	char buffer[BUF_LEN];
	const char *debug;

	if((debug=getenv("DEBUG"))==NULL)
		return;

	va_start(fmtargs,msg);
	vsnprintf(buffer,sizeof(buffer)-1,msg,fmtargs);
	va_end(fmtargs);
	printf("%s",buffer);
}

void log_element(void *data, void *nothing)
{
	if(getenv("DEBUG")==NULL)
		return;
	printf("\nLOG_ELEMENT :  %s", (char *)data);
}


/* 

   Override cupsDoFileRequest so we can remove unwanted printers.

   This function intercepts all requests to the Cups server. For every printer, 
   it checks for the presence of a file named $HOME/.gtklp/accept/<NAME>. If 
   the file doesn't exist, it removes the printer from the Cups request.

   A cups request consists of a series of attributes, grouped together. The
   structure in memory is a list, and each record contains a pointer to the
   next attribute. To hide a printer, we set the pointer so it skips to the
   next printer's first attribute, or NULL if it is the last printer.

*/
ipp_t *cupsDoFileRequest(http_t *http, ipp_t *request, const char *resource, const char *filename)
{
	static ipp_t *(*libcupsDoFileRequest) 
		(http_t *http, ipp_t *request, const char *resource, const char *filename) = NULL;

	void *handle;
	char *error;
	ipp_t *myrequest;

	int  cnt=0;					/* Loop count */
	guint  p=0;					/* Printer count */
	int isprinter=0;				/* attribute is for a printer */
	const char *home;				/* Home directory of the user */
	char printerfilepath[BUF_LEN+IPP_MAX_NAME];	/* Path of printer accept file */
	ipp_attribute_t *attr;                		/* Current attribute */
	GList *prtname = NULL;				/* List of printer names */
	GList *prtlist = NULL;				/* List of pointer to printers */
	ipp_attribute_t *lastattr=NULL;
	struct stat buf;				/* Temporary buffer for stat() */


	if(!libcupsDoFileRequest) {
		handle = dlopen("/usr/lib/libcups.so.2", RTLD_LAZY);

		if (!handle) {
			fputs(dlerror(), stderr);
			exit(1);
		}
		libcupsDoFileRequest = dlsym(handle, "cupsDoFileRequest");
		if ((error = dlerror()) != NULL) {
			fprintf(stderr, "%s\n", error);
			exit(1);
		}
	}

	myrequest = libcupsDoFileRequest(http,request,resource,filename);
	if(request->request.op.operation_id !=CUPS_GET_PRINTERS)
		return(myrequest);
	if((home=getenv("HOME"))==NULL)
	{
		logit("No $HOME!\n");
		return(myrequest);
	}
	else
	{
		logit("Home Directory: %s\n",home);
	}


	/* Get all attributes from the CUPS server */
	for (attr = myrequest->attrs; attr != NULL; attr = attr->next)
	{
		/* It this a printer? */
		if(attr->group_tag==IPP_TAG_PRINTER)
		{
			/* Is it the first tag of a new printer? */
			if(isprinter==0) 
			{  
				p++;
				isprinter=1;
				prtlist = g_list_append(prtlist, attr);
				logit("\nFOUND PRINTER %d! (%p points to it)\n",p,lastattr);
			}
			/* Get the printer name */
			if(attr->name)
			{
				logit("[%d] %s (%p)\n",p,attr->name,attr);
				if (strcmp(attr->name, "printer-name") == 0 &&
						attr->value_tag == IPP_TAG_NAME)
				{
					prtname = g_list_append(prtname, attr->values[0].string.text);
					logit("[%d] ** NAME: %s **\n",p,attr->values[0].string.text);
				}
			}
		}
		else
		{
			/* This is not a printer, but general info from cups */
			isprinter=0;
			logit("[%d] %s (%p)\n",p,attr->name,attr);
		}
		lastattr=attr;
	}

	logit("\n*** End of printers*** (last %p)\n\n",lastattr);
	lastattr=NULL;

	logit("\n*** GLIST PRINTER LIST***\n\n");
	printf("############  BEFORE\n");
	g_list_foreach(prtname, log_element, NULL);
	printf("\n");

	/* List from last to first since we need to know the pointers
	   to the next printers before we can point to it */
	for(cnt=p-1; cnt>=0; cnt--)
	{
		logit("Printer %d = %s\n",cnt,g_list_nth_data(prtname, (guint)cnt));
		strcpy(printerfilepath,home);
		strcat(printerfilepath,"/.gtklp/accept/");
		strcat(printerfilepath,g_list_nth_data(prtname, (guint)cnt));
		logit("Checking for %s\n",printerfilepath);
		if(stat(printerfilepath,&buf)!=0)
		{ 
			logit("DELETE PRINTER %s\n",g_list_nth_data(prtname, (guint)cnt));
			/* The attribute that points to us as the next
			   attribute now points to the next available
			   printer, so this printer is effectively removed
			   from the attribute list */
			((ipp_attribute_t *) g_list_nth_data(prtlist, (guint)cnt))->next = lastattr;
		}
		else
		{
			/* This is a valid printer. Save its pointer */
			lastattr=(ipp_attribute_t *) g_list_nth_data(prtlist, (guint)cnt);
		}
	}

	logit("\n*** GLIST PRINTER LIST***\n\n");
	printf("############  AFTER\n");
	g_list_foreach(prtname, log_element, NULL);

	return myrequest;
}

/* 

   - Override cupsGetDest because if we remove the default printer in
   the previous function, there will be no default printer. Mozilla
   absolutely needs one, otherwise it will segfault.

   Cups uses the PRINTER environment variable by default. If the variable is
   not present, or the printer does not exist, it will use the default 
   printer in Cups. If there is none, it returns NULL.

   We check for this NULL value and replace it instead with the first 
   available printer.

*/

cups_dest_t *cupsGetDest(const char *name, const char *instance, int num_dests, cups_dest_t *dests)    
{
	static cups_dest_t *(*libcupsGetDest)
		(const char *name, const char *instance, int num_dests, cups_dest_t *dests) = NULL;

	void *handle;
	char *error;
	cups_dest_t *mydest;

	if(!libcupsGetDest) {
		handle = dlopen("/usr/lib/libcups.so.2", RTLD_LAZY);

		if (!handle) {
			fputs(dlerror(), stderr);
			exit(1);
		}
		libcupsGetDest = dlsym(handle, "cupsGetDest");
		if ((error = dlerror()) != NULL) {
			fprintf(stderr, "%s\n", error);
			exit(1);
		}
	}

	mydest = libcupsGetDest(name,instance,num_dests,dests);

	if(name==NULL)
	{
		logit("Asking for default printer\n");
		if(mydest==NULL) 
		{
			logit("There is no spoon..\n");
			if(num_dests>0)
			{
				logit("But there are %d printers\n",num_dests);
				logit("Set default to %s\n",dests->name);
				dests->is_default=1;
				return(dests);
			}
		}  
	}
	return(mydest);

}
